#include <SPI.h>
#include <Ethernet.h>
//#include <SD.h>
#include <EEPROM.h>
#include <Bridge.h>
#include <BridgeClient.h>
#include <FileIO.h>
#include <SoftwareSerial.h>
#include <Bounce2.h>

// -- Settings for YOU to change if you want


//byte mac[] = { 0x90, 0xA2, 0xDA, 0xF0, 0x4A, 0x84 }; // physical mac address
//byte ip[] = { 192, 168, 43, 215 };//IP address: 192.168.43.215/255.255.255.0

// The printerType controls the format of the data sent from the server
// If you're using a completely different kind of printer, change this
// to correspond to your printer's PrintProcessor implementation in the
// server.
//
// If you want to control the darkness of your printouts, append a dot and
// a number, e.g. A2-raw.240 (up to a maximum of 255).
//
// If you want to flip the vertical orientation of your printouts, append
// a number and then .flipped, e.g. A2-raw.240.flipped
const char printerType[] = "A2-raw";

const char host[] = "printer.exciting.io"; // the host of the backend server
const unsigned int port = 80;

const unsigned long pollingDelay = 10000; // delay between polling requests (milliseconds)

const byte printer_TX_Pin = 9; // this is the yellow wire
const byte printer_RX_Pin = 8; // this is the green wire
const byte errorLED = 7;       // the red LED
const byte downloadLED = 6;    // the amber LED
const byte readyLED = 5;       // the green LED
const byte buttonPin = 3;      // the print button
const byte SD_Pin = 4;         // the SD Card SPI pin

#define DEBUG // When debug is enabled, log a bunch of stuff to the hardware Serial

// -- Everything below here can be left alone

const char sketchVersion[] = "1.0.6";

// -- Debugging

#ifdef DEBUG
void debugTimeAndSeparator() {
  Serial.print(millis()); Serial.print(": ");
}
void debug(const char *a) {
  debugTimeAndSeparator(); Serial.println(a);
}
#define debug2(a, b) debugTimeAndSeparator(); Serial.print(a); Serial.println(b);
#else
#define debug(a)
#define debug2(a, b)
#endif


// -- Initialize the printer ID

const byte idAddress = 0;
char printerId[17]; // the unique ID for this printer.

inline void initPrinterID() {
  debug("initPrinterID Success");
  if ((EEPROM.read(idAddress) == 255) || (EEPROM.read(idAddress+1) == 255)) {
    debug("Generating new ID");
    randomSeed(analogRead(0) * analogRead(5));
    for(int i = 0; i < 16; i += 2) {
      printerId[i] = random(48, 57); // 0-9
      printerId[i+1] = random(97, 122); // a-z
      EEPROM.write(idAddress + i, printerId[i]);
      EEPROM.write(idAddress + i+1, printerId[i+1]);
    }
  } else {
    for(int i = 0; i < 16; i++) {
      printerId[i] = (char)EEPROM.read(idAddress + i);
    }
  }
  printerId[16] = '\0';
  debug2("ID: ", printerId);
}


// -- Initialize the LEDs

inline void initDiagnosticLEDs() {
  debug("initDiagnosticLEDs Success");
  pinMode(errorLED, OUTPUT);
  pinMode(downloadLED, OUTPUT);
  pinMode(readyLED, OUTPUT);
  digitalWrite(errorLED, HIGH);
  digitalWrite(downloadLED, HIGH);
  digitalWrite(readyLED, HIGH);
  delay(1000);
  digitalWrite(errorLED, LOW);
  digitalWrite(downloadLED, LOW);
  digitalWrite(readyLED, LOW);
  delay(500);
}

// -- Initialize the printer connection

SoftwareSerial *printer;
#define PRINTER_WRITE(b) printer->write(b)

inline void initPrinter() {
  printer = new SoftwareSerial(printer_RX_Pin, printer_TX_Pin);
  printer->begin(9600);
}


// -- Initialize the SD card

inline void initSD() {
  debug("Init SD");
 // pinMode(SD_Pin, OUTPUT);
  //if (!SD.begin(SD_Pin)) {
    // SD Card failure.
    //terminalError(2);
  //}
  
}


// -- Initialize the Ethernet connection & DHCP

//EthernetClient client;
BridgeClient client;
inline void initNetwork() {
   debug("Init Network");
   Bridge.begin();
   delay(1000);
   debug2("IP Address: ", getIPaddr() );
FileSystem.begin();
    /* test that it can write to the SD card */
  File myFile = FileSystem.open("/mnt/sda1/datalog.txt", FILE_WRITE);
  /* File myFile = SD.open("test.txt", FILE_WRITE); */
  if (myFile) {
    debug("Writing to test.txt...");
    char written = myFile.println("testing 1, 2, 3.");
    debug2("wrote: ", written);
  // close the file:
    myFile.close();
    debug("done.");
  } else {
    // if the file didn't open, print an error:
    debug("error opening test.txt");
  }
   
  // start the Ethernet connection:
 // if (Ethernet.begin(mac) == 0) {
    // DHCP Failure
    //debug2("mac", mac);
    //for (byte thisByte = 0; thisByte < 6; thisByte++) {
    // print the value of each byte of the IP address:
    //Serial.print(mac[thisByte], HEX);
    //Serial.print(".");
    //}
    //debug("Error Network");
    
   // terminalError(10);
 // }
  //delay(1000);
  // print your local IP address:
 // debug2("IP: ", Ethernet.localIP());
}


// -- Initialize debouncing of buttons

Bounce bouncer = Bounce();

void initBouncer() {
  bouncer.attach(buttonPin);
  bouncer.interval(5);
}

// -- Setup; runs once on boot.

void setup(){
#ifdef DEBUG
  Serial.begin(9600);
#endif
  debug("Initialization");
  initDiagnosticLEDs();
  initPrinterID();
  initSD();
  initNetwork();
  initPrinter();
  initBouncer();
}

// -- Check for new data and download if found

boolean downloadWaiting = false;
char cacheFilename[] = "/mnt/sda1/TMP";
unsigned long content_length = 0;
boolean statusOk = false;

void checkForDownload() {
  unsigned long length = 0;
  content_length = 0;
  statusOk = false;

#ifdef DEBUG
  unsigned long start = millis();
#endif

  if (FileSystem.exists(cacheFilename)) {
    if (!FileSystem.remove(cacheFilename)) {
      // Failed to clear cache.
      terminalError(4);
    }
  }
  File cache = FileSystem.open(cacheFilename, FILE_WRITE);
  
  debug2("Cache Filename ",cacheFilename);
  //Serial.write(cache.read());
  //debug2("IP Address :",getIPaddr());
  debug2("Attempting to connect to ", host);
  if (client.connect(host, port)) {
    debug("Yeay Connected");
    digitalWrite(downloadLED, HIGH);
    client.print("GET "); client.print("/printer/"); client.print(printerId); client.println(" HTTP/1.0");
    client.print("Host: "); client.print(host); client.print(":"); client.println(port);
    client.flush();
    client.print("Accept: application/vnd.exciting.printer."); client.println(printerType);
    client.print("X-Printer-Version: "); client.println(sketchVersion);
    client.println();
    boolean parsingHeader = true;

    while(client.connected()) {
      while(client.available()) {
        //debug("file available");
        if (parsingHeader) {
          client.find((char*)"HTTP/1.1 ");
          char statusCode[] = "xxx";
          client.readBytes(statusCode, 3);
          statusOk = (strcmp(statusCode, "200") == 0);
          debug2("statusOk :",statusOk);
          client.find((char*)"Content-Length: ");
          char c;
          while (isdigit(c = client.read())) {
            content_length = content_length*10 + (c - '0');
            debug2("calculating content_length :", c);
          }
          debug2("Content length: ", content_length);
          client.find((char*)"\n\r\n"); // the first \r may already have been read above
          parsingHeader = false;
        } else {
            if(cache) {
                debug("masuk");
                cache.write(client.read());
          //cache.close();
          //cache.write(client.read());
          length++;
              }
              else {
                debug("failed cache");
              }
          
        }
      }
      debug("Waiting for data");
    }

    debug("Server disconnected");
    digitalWrite(downloadLED, LOW);
    // Close the connection, and flush any unwritten bytes to the cache.
    client.stop();
    cache.seek(0);

    if (statusOk) {
      if ((content_length == length) && (content_length == cache.size())) {
        debug2("content_length :", content_length);
        if (content_length > 0) {
          downloadWaiting = true;
           debug("part download");
          digitalWrite(readyLED, HIGH);
        }
      }
#ifdef DEBUG
      else {
        debug2("Failure, content length: ", content_length);
        if (content_length != length) debug2("length: ", length);
        if (content_length != cache.size()) debug2("cache: ", cache.size());
        digitalWrite(errorLED, HIGH);
      }
#endif
    } else {
      debug("Response code != 200");
      recoverableError();
    }
  } else {
    debug("Couldn't connect");
    recoverableError();
  }

  cache.close();

#ifdef DEBUG
  unsigned long duration = millis() - start;
  debug2("Bytes: ", length);
  debug2("Duration: ", duration);
#endif
}

void flashErrorLEDs(unsigned int times, unsigned int pause) {
  while (times--) {
    digitalWrite(errorLED, HIGH); delay(pause);
    digitalWrite(errorLED, LOW); delay(pause);
  }
}

inline void recoverableError() {
  flashErrorLEDs(5, 100);
}

inline void terminalError(unsigned int times) {
  flashErrorLEDs(times, 500);
  digitalWrite(errorLED, HIGH);
  // no point in carrying on, so do nothing forevermore:
  while(true);
}

// -- Print send any data from the cache to the printer

inline void printFromDownload() {
  File cache = FileSystem.open(cacheFilename);
  byte b;
  while (content_length--) {
    b = (byte)cache.read();
    PRINTER_WRITE(b);
  }
  cache.close();
  downloadWaiting = false;
  digitalWrite(readyLED, LOW);
}

String getIPaddr() 
{
  Process wifiCheck;  // initialize a new process
  String str, addr;
  int cnt = 10;
  
  while(true)
  {
    str = "";
    
    wifiCheck.runShellCommand("/usr/bin/pretty-wifi-info.lua | grep \"IP address\"");
  
    // while there's any characters coming back from the
    // process, print them to the serial monitor:
    while (wifiCheck.available() > 0) 
    {
      char c = wifiCheck.read();
      str += c;
      Serial.print(c);
    }
  
    Serial.println();
  
    // Trim off "IP address: "
    addr = str.substring(12);
  
    // Trim off the netmask
    int i = addr.indexOf('/');
    addr.remove(i);

    if (addr.length() > 0)
    {
      // Success!
      break;
    }

    if (cnt-- <= 0)
    {
      // Failed
      digitalWrite(13, LOW);
      delay(1000);
      digitalWrite(13, HIGH);
      delay(1000);
      digitalWrite(13, LOW);
      break;
    }
  }

//  Serial.println(addr);
  return addr;
}
// -- Check for new data, print if the button is pressed.

void loop() {

  debug("Inital Loops");
  if (downloadWaiting) {
    bouncer.update();
    if (bouncer.read() == HIGH) {
      printFromDownload();
    }
  } else {
    checkForDownload();
    if (!downloadWaiting) {
      delay(pollingDelay);
    }
  }
}
